#!/usr/bin/tclsh
# OpenVerse Server Program
# 
# This is the server code! :)
#
# Module Name		- Server Program
# Current Maintainter 	- Cruise <cruise@openverse.org>
# Sourced By		- Command Line Or Client
#
# Copyright (C) 1999 David Gale <cruise@openverse.org>
# For more information visit http://OpenVerse.org/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
# USA.

#
# Central Logging Facility
#
# Usage: LogIt "Text To Log"
#
# Everything uses this so we must define it first.
#
# Checked for maximum Memmory Efficiency on 12/09/1999 13:33
#
proc LogIt {text} {
	global MVS

	if $MVS(standalone) {
		# we are running from a command line
		puts $text
	} else {
		# we are running from a gui
		if [winfo exists .ovserver] {
			.ovserver.log insert end "$text\n"
			.ovserver.log see end
			# TODO (6) Check for number of lines and trim.
		}
	}
	unset text
}

#
# Send text to ALL users.
#
# Usage: SendToAllUsers "Text To Send"
#
# This function will send the provided text to all connected users. The
# text should be pre-formated and ready to go.
#
# Checked for maximum memory efficiency 12/09/1999 13:36
#
proc SendToAllUsers {what} {
	global MVS
	
	foreach sckt $MVS(socks) {
		if {$MVS($sckt.name) != "*"} {
			LogIt "($sckt)-> $what"
			catch {puts $sckt "$what"}
		}
	}
	unset what
}

#
# Setup some server variables. Please see the bottom of this document for
# a completel ist of variables which the server will use.
#
global MVS tcl_interactive
set MVS(waiter) 1
set MVS(serving) 1
set MVS(users) 0
set MVS(socks) {}
set MVS(tell_registry) {}
set MVS(entry_registry) {}
set MVS(submit_registry) {}

if !$tcl_interactive {
	set MVS(standalone) 1
	set app "$argv0"
	catch { set app [file readlink $app] }
	set MVS(homedir) "[file dirname $app]"
	set MVS(configfile) "$MVS(homedir)/$argv"
	unset app
	# Sanity Checker.
	#
	# Usage: SanityCheck "pathname"
	#
	# Checks a path to be sure it passed sane rules. It will return 1
	# if this path passes or else it will return 0 if it does not
	# pass.
	#
	# Only defined if running standalone. (command line daemon)
	#
	# Checked for maximum memmory efficiency 12/09/1999 13:59
	#
	proc SanityCheck {what} {
		if {[string first "../" $what] != -1} {
			unset what
			return 0
		elseif {[string first "//" $what] != -1} {
			unset what
			return 0
		elseif {[string first "~/" $what] != -1} {
			unset what
			return 0
		elseif {[string range $what 0 0] == "/"} {
			unset what
			return 0
		} else { 
			unset what
			return 1
		}
	}
	#
	# Test a value to be sure it is a number.
	#
	# Usage: TestNum "12345"
	#
	# This function will test a value to be sure it is a number. The
	# function will return 0 if it is a number and 1 if it is not a
	# number.
	# only defined if running standalone. (command line daemon)
	#
	# Checked for maximum memmory efficiency 12/09/1999 13:59
	#
	proc TestNum {what} {
		global MVS
		set bad 0
		for {set c 0} {$c < [string length $what] && !$bad} {incr c} {
			switch -exact -- [string range $what $c $c] {
				1 {}
				2 {}
				3 {}
				4 {}
				5 {}
				6 {}
				7 {}
				8 {}
				9 {}
				0 {}
				- {}
				default {set bad 1}
			}
		}
		unset c what
		if $bad {
			unset bad
			return 1
		} else {
			unset bad
			return 0
		}
	}
} else {
	set MVS(standalone) 0
	set MVS(homedir) $homedir
	set MVS(configfile) "$MVS(homedir)/pserver.cfg"
	set argv "Sourced!"
}

# (re)Load the server config file.
#
# Usage: ReloadConfig
#
# This function will (re)load the server config file. Any changes will
# take effect at the time of reload.
#
# Checked for maximum memory efficiency 12/09/1999 15:35
#
proc ReloadConfig {} {
	global MVS

	LogIt "------------ Loading Config File -----------"

	set MVS(port) "7000"
	set MVS(timeout) 120
	set MVS(roomname) "My Own Room"
	set MVS(avatars) "$MVS(homedir)/simages"
	set MVS(sobjects) "$MVS(homedir)/sobjects"
	set MVS(tickler) "$MVS(homedir)/TickleMe"
	set MVS(images) "$MVS(homedir)/images"
	set MVS(icons) "$MVS(homedir)/icons"
	set MVS(maxheight) "200"
	set MVS(push) 1
	set MVS(maxwidth) "320"
	set MVS(roomdir) "$MVS(homedir)/rooms"
	set MVS(roomfile) "room.gif"
	set MVS(sendbuffer) 4096
	set MVS(exits) {}
	set MVS(locations) {}
	set MVS(captains) {}
	set MVS(maxpushdistance) 100
	set MVS(maxpushvelocity) 100

	# Create required directories.

	if ![file exists $MVS(avatars)] {file mkdir "$MVS(avatars)"}
	if ![file exists $MVS(sobjects)] {file mkdir "$MVS(sobjects)"}
	if ![file exists $MVS(images)] {file mkdir "$MVS(images)"}
	if ![file exists $MVS(roomdir)] {file mkdir "$MVS(roomdir)"}
	if ![file exists $MVS(icons)] {file mkdir "$MVS(icons)"}

	if [file exists "$MVS(configfile)"] {source $MVS(configfile)}
	if [file exists "$MVS(tickler)"] {
		catch {
			file delete -force "$MVS(tickler)"
		}
	}
	SendToAllUsers "ROOMNAME $MVS(roomname)"
	SendToAllUsers "ROOM $MVS(roomfile) [file size "$MVS(roomdir)/$MVS(roomfile)"]"
}

if {$argv == ""} {
	puts "Usage: server.tcl <Config File>"
	exit
}

ReloadConfig

set MVS(server_socket) [socket -server NewConnect $MVS(port)]
if $tcl_interactive {
	set args [fconfigure $MVS(server_socket) -sockname]
	set MV(MyIpAddress) [lindex $args 0]
}
LogIt "(!) -- SOCKET_OPEN $MVS(server_socket)"

# Accept New Connections.
#
# Usage: NewConnect socket address port
#
# This function will accept a new connection and setup some initial
# variables for the user. It will also set a trigger on the incoming
# socket which will read data on the socket.
#
# checked for maximum memmory efficiency 12/09/1999 18:07
#
proc NewConnect {sck address port} {
	global MVS

	LogIt "($sck)<- New Connection! $address:$port"
	fconfigure $sck -blocking 0 -buffering line
	fileevent $sck readable "Serv_ReadFrom $sck"
	if {[lsearch $MVS(socks) $sck] == -1} {
		lappend MVS(socks) $sck
	} else {
		close $sck
		unset sck address port
		return
	}
	set MVS($sck.name) "*"
	set MVS($sck.address) "$address"
	set MVS($sck.port) "$port"
	set MVS($sck.ping) 0
	set MVS($sck.ping_response) [clock seconds]
	set MVS($sck.x) -1
	set MVS($sck.y) -1
	set MVS($sck.avatar) "*connecting*"
	set MVS($sck.av_head_x) "-1"
	set MVS($sck.av_head_y) "-1"
	set MVS($sck.downloads) {}
	incr MVS(users)
	if !$MVS(standalone) {
		.ovserver.buttons.info.v configure -text $MVS(users)
	}
	unset sck address port
}

#
# Check name validity.
#
# Usage: CheckName "NickName"
#
# this function will check a given nickname to be sure it is allowed. Some
# nicknames are not allowed if they contain special characters.
#
# Checked for maximum memmory efficiency 12/09/1999 18:20
#
proc CheckName {name} {
	if {[string trim $name] == "" || \
		[string trim $name] == "*" || \
		[string trim $name] == "." || \
		[string range $name 0 0] == "-"} {
		unset name
		return 0
	} else {
		unset name
		return 1
	}
}

# Read Incomming Text
# 
# Usage: Serv_ReadFrom socket
#
# This function will read text from a socket and process it. If the user
# is not authenticated, it will authenticate them. If the user is just
# requesting a number of connected users, this function will process the
# request. If it has nothing to do other than to read the text, it will
# pass this text off to the Serv_ProcessInput function for processing.
#
# Checked for maximum memmory efficience on 12/09/1999 18:27
#
proc Serv_ReadFrom {who} {
	global MVS
	
	set input ""
	catch {gets $who input}
	# Update before we process anything!
	if {[eof $who] == 1} {
		if {$MVS($who.name) == "*"} {
			DisconnectUser $who 0
		} else {
			DisconnectUser $who 1
		}
		unset input who
		return
	}

	if {$MVS($who.name) == "*"} {
		puts "<- $input"
		if {[string range $input 0 3] != "AUTH"} {
			if {$input == "USERS"} {
				SendToUser $who "USERS [expr [llength $MVS(socks)] -1]"
				DisconnectUser $who 0
			} else {
				SendToUser $who "AUTH REQD"
				DisconnectUser $who 0
			}
		} else {
			set parms [split [string range $input 5 end] " "]
			if {[TestNum [lindex $parms 1]] || \
			  [TestNum [lindex $parms 2]] || \
			  [TestNum [lindex $parms 4]] || \
			  [TestNum [lindex $parms 5]] || \
			  [TestNum [lindex $parms 6]] || \
			  [TestNum [lindex $parms 7]] || \
			  [TestNum [lindex $parms 8]] } {
				SendToUser $who "AUTH FAILED (Non Numeric)"
				DisconnectUser $who 0
				unset input who
				return
			}
			if {[string length [lindex $parms 1]] > 3 || \
				[string length [lindex $parms 2]] > 3 || \
				[string length [lindex $parms 4]] > 4 || \
				[string length [lindex $parms 5]] > 4 || \
				[string length [lindex $parms 6]] > 6 || \
				[string length [lindex $parms 7]] > 4 || \
				[string length [lindex $parms 8]] > 4 || \
				[lindex $parms 1] < 0 || \
				[lindex $parms 2] < 0 || \
				[lindex $parms 6] < 0} {
				SendToUser $who "AUTH FAILED (String Lengths)"
				DisconnectUser $who 0
				unset input who
				return
			}

			set MVS($who.name) [string range [lindex $parms 0] 0 12]
			set MVS($who.x) [lindex $parms 1]
			set MVS($who.y) [lindex $parms 2]
			set MVS($who.avatar) [lindex $parms 3]
			set MVS($who.av_head_x) [lindex $parms 4]
			set MVS($who.av_head_y) [lindex $parms 5]
			set MVS($who.av_baloon_x) [lindex $parms 7]
			set MVS($who.av_baloon_y) [lindex $parms 8]
			set size [lindex $parms 6]

			if ![CheckName $MVS($who.name)] {
					SendToUser $who "BADNAME"
					DisconnectUser $who 0
					unset input parms size who
					return
			}
			foreach sckt $MVS(socks) {
				if {$MVS($sckt.name) == $MVS($who.name) && $sckt != $who} {
					SendToUser $who "NAMEINUSE"
					DisconnectUser $who 0
					unset input parms size sckt who
					return
				}
			}

			SendToUser $who "ROOMNAME $MVS(roomname)"
			SendToUser $who "ROOM $MVS(roomfile) [file size $MVS(roomdir)/$MVS(roomfile)]"
			SendObjects $who

			if ![file exists $MVS(avatars)/$MVS($who.avatar)] {
				SendToAllUsers "NEW $MVS($who.name) $MVS($who.x) $MVS($who.y) default.gif 0 20 [file size $MVS(icons)/default.gif] 24 6"
				LogIt "($who) $MVS($who.avatar) does not exist"
				GetBinaryFile $who $MVS($who.avatar) $size
			} else {
				if {[file size $MVS(avatars)/$MVS($who.avatar)] != $size} {
					SendToAllUsers "NEW $MVS($who.name) $MVS($who.x) $MVS($who.y) default.gif 0 20 [file size $MVS(icons)/default.gif] 24 6"
					LogIt "($who) $MVS($who.avatar) $size != [file size $MVS(avatars)/$MVS($who.avatar)]"
					GetBinaryFile $who $MVS($who.avatar) $size
				} else {
					if [CheckGif "$MVS(avatars)/$MVS($who.avatar)"] {
						SendToAllUsers "NEW $MVS($who.name) $MVS($who.x) $MVS($who.y) $MVS($who.avatar) $MVS($who.av_head_x) $MVS($who.av_head_y) [file size $MVS(avatars)/$MVS($who.avatar)] $MVS($who.av_baloon_x) $MVS($who.av_baloon_y)"
					} else {
						SendToAllUsers "NEW $MVS($who.name) $MVS($who.x) $MVS($who.y) default.gif 0 20 [file size $MVS(icons)/default.gif] 24 6"
						SendToUser $who "TOOBIG"
						set MVS($who.avatar) "default.gif"
					}
				}
			}

			foreach sckt $MVS(socks) {
				if {$MVS($sckt.name) != $MVS($who.name)} {
					if [file exists $MVS(avatars)/$MVS($sckt.avatar)] {
						SendToUser $who "NEW $MVS($sckt.name) $MVS($sckt.x) $MVS($sckt.y) $MVS($sckt.avatar) $MVS($sckt.av_head_x) $MVS($sckt.av_head_y) [file size $MVS(avatars)/$MVS($sckt.avatar)] $MVS($sckt.av_baloon_x) $MVS($sckt.av_baloon_y)"
					} else {
						SendToUser $who "NEW $MVS($sckt.name) $MVS($sckt.x) $MVS($sckt.y) default.gif 0 20 [file size $MVS(icons)/default.gif] 24 6"
					}
				}
			}
		}
		unset input who
		return
	}
	Serv_ProcessInput $who $input
	unset input who
}

# Disconnect users from the system.
#
# Usage: DisconnectUser socket announce_disconnect
#
# This function is used to disconnect a user from the system.
# A general cleanup will be done with the variables the user was
# consuming and an annoucement will be made to all connected users if it
# is requested with the announce_disconnect parameter (set to 1)
#
# Checked for maximum memmory efficiency 12/09/1999 19:20
#
proc DisconnectUser {who announce} {
	global MVS
	
	if {[lsearch -exact $MVS(socks) $who] == -1} {return}

	LogIt "($who)<- Disconnected! $MVS($who.address):$MVS($who.port)"
	incr MVS(users) -1
	if !$MVS(standalone) {
		.ovserver.buttons.info.v configure -text $MVS(users)
	}
	catch {close $who}
	LogIt "(!) -- SOCKET_CLOSE $who"
	set which [lsearch -exact $MVS(socks) $who]
	set MVS(socks) [lreplace $MVS(socks) $which $which]

	if $announce {
		foreach sckt $MVS(socks) {
			SendToUser $sckt "NOMORE $MVS($who.name)"
		}
	}
	#
	# Clean up the mess this user made!
	#
	catch {
		uplevel #0 unset MVS($who.name) MVS($who.address) MVS($who.port) \
			MVS($who.ping) MVS($who.ping_response) MVS($who.x) \
			MVS($who.y) MVS($who.avatar) MVS($who.av_head_x) \
			MVS($who.av_head_y) MVS($who.av_baloon_x) \
			MVS($who.av_baloon_y) MVS($who.downloads)
	}
	#
	# TODO (7) Downloads in progress should be canceled!
	#
	unset which who announce
}

# Send Text To a Connect User
#
# Usage: SendToUser socket "text to send"
#
# This function will send the provided text to the user specified
# It should be pre-formated and ready to go.
#
proc SendToUser {who what} {
	LogIt "($who)-> $what"
	catch {puts $who "$what"}
	unset who what
}

# Process Input from users.
#
# Usage: Serv_ProcessInput socket "text to process"
#
# This function is the root of the protocol. It processes all of the
# things which a client can send to the server. If the client sends
# something it does not understand... it will ignore it. Please see the
# protocol documentation within the technical documentation for a complete
# descritption of the logic within this function.
#
# Checked for maximum memmory efficiency 12/09/1999 22:05
#
proc Serv_ProcessInput {who what} {
	global MVS

	LogIt "($who)<- $what"
	if {[string first " " $what] != -1} {
		set cmd [string range $what 0 [expr [string first " " $what] -1]]
		set rest [string range $what [expr [string first " " $what] +1] end]
		set parms [split $rest " "]
	} else {
		set cmd $what
		set rest ""
		set parms {}
	}
	switch -exact -- $cmd {
		"MOVE" {
			#
			#fixes a bug that MVSbMVS found.
			#
			set retflag 0
			if {[TestNum [lindex $parms 1]] || \
				[TestNum [lindex $parms 2]] || \
				[TestNum [lindex $parms 3]]} {set retflag 1}
			if {[string length [lindex $parms 1]] > 4} {set retflag 1}
			if {[string length [lindex $parms 2]] > 4} {set retflag 1}
			if {[string length [lindex $parms 3]] > 2} {set retflag 1}
			if {[lindex $parms 1] < 0} {set retflag 1}
			if {[lindex $parms 2] < 0} {set retflag 1}
			if {[lindex $parms 3] < 0} {set retflag 1}
			if $retflag {
				unset cmd rest parms retflag
				return
			}
			set MVS($who.x) [lindex $parms 1]
			set MVS($who.y) [lindex $parms 2]
			set is_exiting 0
			set idx 0
			foreach exit $MVS(exits) {
				set exl [split $exit " "]
				set x1 	[lindex $exl 0]
				set y1 	[lindex $exl 1]
				set x2 	[lindex $exl 2]
				set y2 	[lindex $exl 3]
				if {$MVS($who.x) > $x1 && \
					$MVS($who.x) < $x2 && \
					$MVS($who.y) > $y1 && \
					$MVS($who.y) < $y2} {
						set is_exiting 1
						set eidx $idx
				}
				incr idx
				unset exl x1 y1 x2 y2
			}
			if $is_exiting {				
				SendToAllUsers "MOVE $MVS($who.name) [lindex $parms 1] [lindex $parms 2] [lindex $parms 3]"
				SendToUser $who "EXIT [lindex $MVS(locations) $eidx]"
				unset eidx
			} else {
				SendToAllUsers "MOVE $MVS($who.name) [lindex $parms 1] [lindex $parms 2] [lindex $parms 3]"
			}
			unset idx is_exiting retflag
		}
		"QUERY" {
			switch -- [lindex $parms 0] {
				"POS_ALL" {
					foreach person $MVS(socks) {
						SendToUser $who "MOVE $MVS($person.name) $MVS($person.x) $MVS($person.y) 50"
					}
				}
			}
		}
		"PUSH" {
			if !$MVS(push) {return}
			set retflag 0
			if {[lindex $parms 0] == ""} {set retflag 1}
			set velocity [lindex $parms 0]
			if {[TestNum $velocity]} {set retflag 1}
			if $retflag {
				unset cmd rest parms retflag velocity
				return
			}
			if { $velocity > $MVS(maxpushvelocity)} {set velocity $MVS(maxpushvelocity)}
			if { $velocity < 0} {set velocity 1}
			foreach s $MVS(socks) {
				if {$who == $s} {continue}
				if {$MVS($who.x) >= $MVS($s.x)} {
					set xdistance [expr $MVS($who.x) - $MVS($s.x)]
					set xpush "-"
				} else {
					set xdistance [expr $MVS($s.x) - $MVS($who.x)]
					set xpush "+"
				}
				if {$MVS($who.y) >= $MVS($s.y)} {
					set ydistance [expr $MVS($who.y) - $MVS($s.y)]
					set ypush "-"
				} else {
					set ydistance [expr $MVS($s.y) - $MVS($who.y)]
					set ypush "+"
				}
				if {$xdistance <= $MVS(maxpushdistance) && $ydistance <= $MVS(maxpushdistance)} {
					LogIt "(PUSH) $xdistance $ydistance $xpush $ypush $velocity"
					set MVS($s.x) [expr $MVS($s.x) $xpush ($velocity - $ydistance)]
					set MVS($s.y) [expr $MVS($s.y) $ypush ($velocity - $xdistance)]
					if {$MVS($s.x) <= 0 } {set MVS($s.x) 10}
					if {$MVS($s.y) <= 0 } {set MVS($s.y) 10}
					if {$MVS($s.x) >= 640 } {set MVS($s.x) 630}
					if {$MVS($s.y) >= 480 } {set MVS($s.y) 470}
					set is_exiting 0
					set idx 0
					foreach exit $MVS(exits) {
						set exl [split $exit " "]
						set x1 	[lindex $exl 0]
						set y1 	[lindex $exl 1]
						set x2 	[lindex $exl 2]
						set y2 	[lindex $exl 3]
						if {$MVS($s.x) > $x1 && \
							$MVS($s.x) < $x2 && \
							$MVS($s.y) > $y1 && \
							$MVS($s.y) < $y2} {
							set is_exiting 1
							set eidx $idx
						}
						incr idx
					}
					catch {unset exl}
					if $is_exiting {
						SendToUser $s "PUSH $MVS($s.x) $MVS($s.y) 20"
						SendToAllUsers "MOVE $MVS($s.name) $MVS($s.x) $MVS($s.y) 20"
						SendToUser $s "EXIT [lindex $MVS(locations) $eidx]"
					} else {
						SendToUser $s "PUSH $MVS($s.x) $MVS($s.y) 20"
						SendToAllUsers "MOVE $MVS($s.name) $MVS($s.x) $MVS($s.y) 20"
					}
				}
			}
			catch {
				unset retflag velocity xdistance xpush ydistance \
				ypush is_exiting idx x1 y1 x2 y2 eidx
			}
		}
		"SEND" {
			if ![SanityCheck [lindex $parms 0]] {return}
			SendBinaryFile $who [lindex $parms 0]
		}
		"DCCSENDAV" {
			if ![SanityCheck [lindex $parms 0]] {return}
			Serv_DCCSend $who [lindex $parms 0] AVATAR
		}
		"DCCSENDOB" {
			if ![SanityCheck [lindex $parms 0]] {return}
			Serv_DCCSend $who [lindex $parms 0] OBJECT
		}
		"DCCSENDROOM" {
			if ![SanityCheck [lindex $parms 0]] {return}
			Serv_DCCSend $who [lindex $parms 0] ROOM
		}
		"EFFECT" {
			SendToAllUsers "EFFECT $MVS($who.name) [lindex $parms 0]"
		}
		"USERS" {
			SendToUser $what "USERS [llength $MVS(socks)]"
		}
		"SUB" {
			if {[lindex $parms 0] == ""} {return}
			set stuff [string range $rest [expr [string first " " $rest] +1] end]
			if {[string length $stuff] > 256} {
				set stuff [string range $stuff 0 256]
			}
			foreach s $MVS(socks) {
				if {$MVS($s.name) == [lindex $parms 0] || [lindex $parms 0] == "*"} {
					SendToUser $s "SUB $MVS($who.name) $stuff"
				}
			}
			unset stuff
		}
		"URL" {
			if {[lindex $parms 0] == ""} {return}
			set stuff [string range $rest [expr [string first " " $rest] +1] end]
			if {[string length $stuff] > 256} {
				set stuff [string range $stuff 0 256]
			}
			foreach s $MVS(socks) {
				if {$MVS($s.name) == [lindex $parms 0] || [lindex $parms 0] == "*"} {
					SendToUser $s "URL $MVS($who.name) $stuff"
				}
			}
			unset stuff
		}
		"PONG" {
			set MVS($who.ping_response) [clock seconds]
		}
		"RSEND" {
			if ![SanityCheck [lindex $parms 0]] {return}
			SendRoomFile $who [lindex $parms 0]
		}
		"CHAT" {
			if {$rest != ""} {
				SendToAllUsers "CHAT $MVS($who.name) [string range $rest 0 256]"
			}
		}
		"SCHAT" {
			set parms [split $rest " "]
			set rest [string range $rest [expr [string first " " $rest] +1] end]
			if {$rest != ""} {
				SendToAllUsers "SCHAT [lindex $parms 0] $MVS($who.name) [string range $rest 0 256]"
			}
		}
		"AVATAR" {
			  if {[TestNum [lindex $parms 1]] || \
			 	[TestNum [lindex $parms 2]] || \
			 	[TestNum [lindex $parms 3]] || \
			 	[TestNum [lindex $parms 4]] || \
			  	[TestNum [lindex $parms 5]]} {
				SendToUser $who "BAD AVATAR"
				unset cmd rest parms
				return
			}
			set retflag 0
			if {[string length [lindex $parms 1]] > 4} {set retflag 1}
			if {[string length [lindex $parms 2]] > 4} {set retflag 1}
			if {[string length [lindex $parms 3]] > 6} {set retflag 1}
			if {[string length [lindex $parms 4]] > 4} {set retflag 1}
			if {[string length [lindex $parms 5]] > 4} {set retflag 1}
			if {[lindex $parms 3] < 0} {set retflag 1}
			if $retflag {
				unset cmd rest parms retflag
				return
			}
			Serv_ChangeAvatar $who [lindex $parms 0] [lindex $parms 1] [lindex $parms 2] [lindex $parms 3] [lindex $parms 4] [lindex $parms 5]
		}
		"WHOIS" {
			if {[lindex $parms 0] == ""} {return}
			foreach s $MVS(socks) {
				if {$MVS($s.name) == [lindex $parms 0] || [lindex $parms 0] == "*"} {
					SendToUser $who "WHOIS $MVS($s.name) $MVS($s.name)@$MVS($s.address)"
				}
			}
		}
		"PRIVMSG" {
			if {[lindex $parms 0] == ""} {return}
			set stuff [string range $rest [expr [string first " " $rest] +1] end]
			if {[string length $stuff] > 256} {
				set stuff [string range $stuff 0 256]
			}
			foreach s $MVS(socks) {
				if {$MVS($s.name) == [lindex $parms 0] || [lindex $parms 0] == "*"} {
					SendToUser $s "PRIVMSG $MVS($who.name) $stuff"
				}
			}
			unset stuff
		}
		"NICK" {
			if {$MVS($who.name) == [lindex $parms 0] || $MVS($who.name) == "*"} {
				unset cmd rest parms
				return
			}
			if ![CheckName [lindex $parms 0]] {
				SendToUser $who "NAMEINUSE"
				unset cmd rest parms
				return
			}
			foreach sckt $MVS(socks) {
				if {$MVS($sckt.name) == [lindex $parms 0]} {
					SendToUser $who "NAMEINUSE"
					DisconnectUser $what 1
					unset cmd rest parms
					return
				}
			}
			SendToAllUsers "NOMORE $MVS($who.name)"
			set MVS($who.name) [string range [lindex $parms 0] 0 12]
			SendToAllUsers "NEW $MVS($who.name) $MVS($who.x) $MVS($who.y) default.gif 0 20 [file size $MVS(icons)/default.gif] 24 6"
 			SendToAllUsers "AVATAR $MVS($who.name) $MVS($who.avatar) $MVS($who.av_head_x) $MVS($who.av_head_y) [file size $MVS(avatars)/$MVS($who.avatar)] $MVS($who.av_baloon_x) $MVS($who.av_baloon_y)"
		}
		"TELL" {
			# Object Interaction. (tell)
			foreach tell $MVS(tell_registry) {
				set reg [split $tell " "]
				if {[lindex $parms 0] == [lindex $reg 0]} {
					[lindex $reg 1] $who
				}
			}
		}
		"SUBMIT" {
			# Object Interaction. (submit)
			foreach submit $MVS(submit_registry) {
				set reg [split $submit " "]
				if {[lindex $parms 0] == [lindex $reg 0]} {
					[lindex $reg 1] $who
				}
			}
		}
		"ENTRY" {
			# Object Interaction.
			set text [string range $rest [expr [string first " " $rest] +1] end]
			if {$text == $rest} {set text ""}
			foreach entry $MVS(entry_registry) {
				set reg [split $entry " "]
				if {[lindex $parms 0] == [lindex $reg 0]} {
					[lindex $reg 1] $who $text
				}
			}
		}
	}
	unset cmd rest parms
}

#
# Change a user's avatar.
#
# Usage: Serv_ChangeAvatar who avatar_name nametag_x nametag_y size
#		balloon_x balloon_y
#
# This function is used to change a user's avatar. It will announce the
# change to all connected users.
#
# Checked for maximum memory efficiency 12/09/1999 22:14
#
proc Serv_ChangeAvatar {who what x y size bx by} {
	global MVS

	set MVS($who.avatar) $what
	set MVS($who.av_head_x) $x
	set MVS($who.av_head_y) $y
	set MVS($who.av_baloon_x) $bx
	set MVS($who.av_baloon_y) $by
	if {$what == "default.gif"} {
		SendToAllUsers "AVATAR $MVS($who.name) $what $MVS($who.av_head_x) $MVS($who.av_head_y) [file size $MVS(icons)/$what] $bx $by"
		return
	}
	if ![file exists $MVS(avatars)/$what] {
		LogIt "-> $what does not exist"
		GetBinaryFile $who $MVS($who.avatar) $size
	} else {
		if {[file size $MVS(avatars)/$what] != $size} {
			LogIt "-> $what $size != [file size $MVS(avatars)/$what]"
			GetBinaryFile $who $MVS($who.avatar) $size
		} else {
			if [CheckGif "$MVS(avatars)/$what"] {
				SendToAllUsers "AVATAR $MVS($who.name) $what $MVS($who.av_head_x) $MVS($who.av_head_y) [file size $MVS(avatars)/$what] $bx $by"
			} else {
				SendToAllUsers "AVATAR $MVS($who.name) default.gif 0 20 [file size $MVS(icons)/default.gif] 24 6"
				SendToUser $who "TOOBIG"
				set MVS($who.avatar) "default.gif"
			}
		}
	}
}

#
# Check various timeouts and events.
#
# Usage: Serv_CheckTimeouts
#
# This function should be run only ONCE when the server is started. It
# should never be run multiple times within a given instance. Once it is
# run, it will re-spawn iteself later and check again, add infinitium.
# What it does is check timeouts. Download Timeouts, Ping Timeouts and it
# also checks for the existance of a tickler file which, when it exists,
# will cause the server to reload it's config file.
#
# Checked for maximum memmory efficiency 12/09/1999 23:31
#
proc Serv_CheckTimeouts {} {
	global MVS

	#
	# Check user ping times.
	#
	set tme [clock seconds]
	set disco {}
	foreach who $MVS(socks) {
		if {[expr $tme - $MVS($who.ping_response)] > 320} {
			LogIt "($who) - Ping Timeout!"
			lappend disco $who
		} else {
			if {[expr $tme - $MVS($who.ping)] > 150} {
				set MVS($who.ping) [clock seconds]
				SendToUser $who "PING"
			}
		}
	}
	foreach who $disco {
		DisconnectUser $who 1
	}
	#
	# Check active downloads.
	#
	set tme [clock seconds]
	foreach idx $MVS(dcc_list) {
		if {[expr $tme - $MVS(DCC.$idx.time)] > $MVS(timeout)} {
			if {$MVS(DCC.$idx.server) > 0} {
				catch {close $MVS(DCC.$idx.server)}
			}
			Serv_endDCC Timer $idx 0 "Connection Timed Out $MVS(DCC.$idx.file)"
		}
	}
	#
	# Check for tickler file
	#
	if [file exists "$MVS(tickler)"] {ReloadConfig}
	#
	# If we're serving, reload this function in 5 seconds.
	#
	if $MVS(serving) {
		after 5000 Serv_CheckTimeouts
	}

	unset disco tme
}

# Error Logging Routine.
#
# Usage: Used internally when an error occurs.
#
# Will print out information on errors
# and continue running (we hope)
#
# checked for maximum memory efficiency 12/10/1999 08:12
#
proc bgerror {stuff} {
	global errorInfo
	global errorCode
	
	LogIt "-------------------------------------"
	LogIt "BGERROR Begin"
	LogIt "-------------------------------------"
	LogIt "Error Code: $stuff"
	LogIt "-------------------------------------"
	LogIt $errorInfo
	LogIt "-------------------------------------"
	LogIt "BGERROR End"
	LogIt "-------------------------------------"
}

#--------------------------------------------
# --           NEW DOWNLOAD CODE!          --
#--------------------------------------------
set MVS(dcc_list) {}
set MVS(dcc_num) 0

# Check for duplicate downloads.
#
# Usage: MVSerifyAvailable socket filename
#
# This function will check to see if the
# user is already getting the file named.
#
# checked for maximum memmory efficiency 12/10/1999 08:15
#
proc MVSerifyAvailable {who what} {
	global MVS


	foreach idx $MVS(dcc_list) {
		if {$MVS(DCC.$idx.sender) == $who && "[file tail $MVS(DCC.$idx.file)]" == $what} {
			return 0
		}
	}
	return 1
}

#
# Send a file
#
# Usage: Serv_DCCSend socket filename transfer_type
#
# This function will initiate a file transfer. It is used for rooms and
# avatars and object images. This is a passive dcc transfer so things work
# better through modern firewalls.
#
# Checked for maximum memmory efficiency 12/10/1999 21:38
#
proc Serv_DCCSend {who what type} {
	global MVS

	if ![MVSerifyAvailable $who $what] {
		LogIt "($who) -- Already Getting $what"
		return
	}

	switch -exact -- $type {
		"AVATAR" {
			if {$what == "default.gif"} {
				set file "$MVS(icons)/default.gif"
			} else {
				set file "$MVS(avatars)/$what"
			}
			set GETCMD "DCCGETAV"
		}
		"OBJECT" {
			set file "$MVS(images)/$what"
			set GETCMD "DCCGETOB"
		}
		"ROOM" {
			set file "$MVS(roomdir)/$what"
			set GETCMD "DCCGETROOM"
		}
		default {return}
	}

	if {[file exists $file]} {
		if {![file readable $file]} {
            		LogIt "($who) (Serv_DCCSend) Cannot read file $file."
			return
		}

		set size [file size $file]
		set idx [incr MVS(dcc_num)]
		set sock [socket -server "Serv_acceptSend $idx" 0]

		if {[catch {fconfigure $sock -sockname} port]} {
			LogIt "($who) (Serv_DCCSend) Cannot get port for server - $port"
		}

		lappend MVS(dcc_list) $idx
		set MVS(DCC.$idx.sender) $who
		set MVS(DCC.$idx.file) "$file"
		set MVS(DCC.$idx.size) $size
		set MVS(DCC.$idx.posn) 0
		set MVS(DCC.$idx.time) [clock seconds]
		set MVS(DCC.$idx.server) $sock
		set MVS(DCC.$idx.sock) -1
		set MVS(DCC.$idx.port) [lindex $port 2]
		set MVS(DCC.$idx.remote) "0.0.0.0"
		set MVS(DCC.$idx.av_head_x) 0
		set MVS(DCC.$idx.av_head_y) 0
		set MVS(DCC.$idx.av_baloon_x) 0
		set MVS(DCC.$idx.av_baloon_y) 0

		SendToUser $who "$GETCMD [lindex $port 2] $what $size"
		unset size idx sock
	} else {
		LogIt "($who) (Serv_DCCSend) File $file does not exist."
	}
	unset file GETCMD
}

#
# Accept pending outbout transfers.
#
# Usage: Serv_acceptSend index socket host port
#
# This function is not called directly, it is instead called by the
# opening of the server socket. It accepts the connection, closes the
# server socket and starts the sending of the file to the user.
#
# checked for maximum memmory efficiency 12/10/1999 21:40
#
proc Serv_acceptSend {index chan hst port} {
	global MVS

	catch {close $MVS(DCC.$index.server)}
	uplevel set MVS(DCC.$index.server) -1
	uplevel #0 set MVS(DCC.$index.sock) $chan
	
	set msg ""

	if {[ catch {open $MVS(DCC.$index.file) RDONLY} infile]} {
		Serv_endDCC Send $index 0 "Cannot read $MVS(DCC.$index.file) : $infile"
		unset infile
		return 0
	}

	if {[set posn $MVS(DCC.$index.posn)] != {} && $posn > 0} {
		if {[catch {seek $infile $posn start} msg]} {
			Serv_endDCC Send $index 0 "Cannot seek $MVS(DCC.$index.file) : $msg"
			close $infile
			unset infile posn msg
			return 0
		}
		uplevel #0 incr MVS(DCC.$index.size) -$posn
	} 

	if {$MVS(DCC.$index.size) == 0} {
		close $infile
		after 50 "Serv_endDCC Send $index 1 \"Transfer completed.\""
		unset infile posn msg
		return 1
	}

	set st [clock seconds]
	fconfigure $infile -translation binary

	if {[catch {set buffer [read $infile $MVS(sendbuffer)]} msg]} {
		Serv_endDCC Send $index 0 "Error reading $file : $msg"
		close $infile
		unset infile posn msg st buffer
		return 0
	}

	global tl
	set tl($chan) [string length $buffer]
	fconfigure $chan -blocking 0 -buffering none -translation binary
	if {[catch {puts -nonewline $chan $buffer} msg]} {
		Serv_endDCC Send $index 0 "Write error : $msg"
		close $infile
		unset infile posn msg st buffer
		return 0
	}
	LogIt "($MVS(DCC.$index.sender)) -- Accepted Serv_DCCSend"
	fileevent $chan readable "Serv_dccSendEvent $index $st $infile"
	unset infile posn msg st buffer
}

#
# The file send event (read trigger) for sending files.
#
# Usage: Serv_dccSendEvent index start_time file_stream_descriptor 
#
# This function is triggered each time the remote client sends a response
# announcing the number of bytes the server has sent to it. It will send
# some more bytes if the client has received all of what we've sent so
# far. It will end the transfer when the client informs us that it has all
# of the data we've sent.
#
# checked for maximum memmory efficiency 12/10/1999 21:47
#
proc Serv_dccSendEvent {index st fd} {
	global MVS
	
	set sk $MVS(DCC.$index.sock)
	uplevel #0 set MVS(DCC.$index.time) [clock seconds]
	set msg ""

	if {[eof $sk]} {
		after 50 "Serv_endDCC Send $index 0 \"Transfer interrupted\""
		close $fd
		unset sk
		return
	}

	if {[catch {set l [read $sk 4]} msg]} {
		Serv_endDCC Send $index 0 "Read error : $msg"
		unset sk l msg
		close $fd
		return
	}

	if {[string length $l] == 0} {
		Serv_endDCC Send $index 0 "Sync read error"
		unset sk l msg
		close $fd
		return
	}

        global tl
	set cl 0
	binary scan $l I1 cl
	if {$cl != $tl($sk)} {
		unset sk l msg cl
		return
	}

	LogIt "($MVS(DCC.$index.sender) Got Check (OK)"
	if [eof $fd] {
		if {[set st [expr {[clock seconds] - $st}]] == 0} {
			    set st 1
		}
		unset sk l msg cl st
		close $fd		
		after 50 "Serv_endDCC Send $index 1 \"Transfer completed\""
		return
	}

	LogIt "($MVS(DCC.$index.sender) Got Check (OK Not EOF)"

	if {[catch {set buffer [read $fd $MVS(sendbuffer)]} msg]} {
		Serv_endDCC Send $index 0 "Error reading $MVS(DCC.$index.file) : $msg"
		unset sk l msg cl st buffer
		close $fd
		return
	}

	if {[set lng [string length $buffer]] == 0} {
		if {[set st [expr {[clock seconds] - $st}]] == 0} {
			set st 1
		}
		unset sk l msg cl st buffer lng
		close $fd
		after 50 "Serv_endDCC Send $index 1 \"Transfer completed.\""
		return
	}
	incr tl($sk) $lng
	LogIt "($MVS(DCC.$index.sender)) -- Sent $lng bytes ($tl($sk) total)"
	if {[catch {puts -nonewline $sk $buffer} msg]} {
		Serv_endDCC Send $index 0 "Write error : $msg"
		unset sk l msg cl st buffer lng
		close $fd
		return
	}

	if {[set dt [expr {[clock seconds] - $st}]] == 0} {
		set elt 0
	} {
		set elt [expr {($MVS(DCC.$index.size) - $tl($sk)) / ($tl($sk) /([clock seconds] - $st))}]
	}
	unset sk l msg cl st buffer lng dt elt
}

# DCC Get Code (Passive)
#
# Usage: GetBinaryFile socket filename file size
#
# This function initializes a passive DCC transfer between a connected
# user and the server. It will set up the listening connection and tell
# the user where to connect to. It will then set up an event to accep the
# connection and transfer the file.
#
# This is our central location for DOWNLOADING files.
# 
# checked for maximum memmory efficiency 12/10/1999 22:28
#
proc GetBinaryFile {who what size} {
	global MVS
	
	
	if ![SanityCheck "$what"] {
		LogIt "(!) $what fails SanityCheck"
		return
	}
	if ![MVSerifyAvailable $who $what] {
		LogIt "($who) -- Already GetTing $what from this user"
		return
	}

	set file "$MVS(avatars)/$what"	
	
	set idx [incr MVS(dcc_num)]
	set sock [socket -server "acceptGet $idx" 0]
	if {[catch {fconfigure $sock -sockname} port]} {
		LogIt "($who) (Serv_DCCSend) Cannot get port for server - $port"
	}
	lappend MVS(dcc_list) $idx
	set MVS(DCC.$idx.sender) $who
	set MVS(DCC.$idx.file) "$file"
	set MVS(DCC.$idx.size) $size
	set MVS(DCC.$idx.posn) 0
	set MVS(DCC.$idx.time) [clock seconds]
	set MVS(DCC.$idx.server) $sock
	set MVS(DCC.$idx.sock) -1
	set MVS(DCC.$idx.port) [lindex $port 2]
	set MVS(DCC.$idx.remote) $MVS($who.address)
	set MVS(DCC.$idx.av_head_x) $MVS($who.av_head_x)
	set MVS(DCC.$idx.av_head_y) $MVS($who.av_head_y)
	set MVS(DCC.$idx.av_baloon_x) $MVS($who.av_baloon_x)
	set MVS(DCC.$idx.av_baloon_y) $MVS($who.av_baloon_y)
	SendToUser $who "DCCSENDAV [lindex $port 2] $what"
	unset file idx sock port
}	

#
# Accept pending inbound transfers.
#
# Usage: Serv_acceptGet index socket host port
#
# This function is not called directly, it is instead called by the
# opening of the server socket. It accepts the connection, closes the
# server socket and starts the getting of the file to the user.
#
# checked for maximum memmory efficiency 12/10/1999 22:35
#
proc acceptGet {index chan hst port} {
	global MVS

	catch {close $MVS(DCC.$index.server)}
	uplevel set MVS(DCC.$index.server) -1
	uplevel #0 set MVS(DCC.$index.sock) $chan

	set file $MVS(DCC.$index.file)
	set posn $MVS(DCC.$index.posn)
	fconfigure $MVS(DCC.$index.sock) -buffering none -blocking 0 -translation binary -buffersize 4096
	set flags [list WRONLY CREAT]
	set msg ""
	if {$posn == 0} { lappend flags TRUNC }
	if {![catch {open $file $flags 0600} outfile]} {
		if {$posn != 0} {
			if {[catch {seek $outfile $posn start} msg]} {
				close $outfile
				Serv_endDCC Get $index 0 "Cannot seek on $file : $msg"
				unset file posn flags outfile msg
				return 0
			}
			uplevel #0 incr MVS(DCC.$index.size) -$posn
		}
		uplevel #0 set tl($MVS(DCC.$index.sock)) 0
		fconfigure $outfile -translation binary 
		fileevent $MVS(DCC.$index.sock) readable "Serv_dccgevent $index [clock seconds] $outfile"
	} {
		Serv_endDCC Get $index 0 "Cannot write $file : $outfile"
		unset file posn flags outfile msg
        	return 0
	}
	unset file posn flags outfile msg
	return 1
}


#
# The file get event (read trigger) for getting files.
#
# Usage: Serv_dccgevent index start_time file_stream_descriptor 
#
# This function is triggered each time the remote client sends a response
# announcing the number of bytes the server has sent to it. It will send
# some more bytes if the client has received all of what we've sent so
# far. It will end the transfer when the client informs us that it has all
# of the data we've sent.
#
# checked for maximum memmory efficiency 12/10/1999 21:47
#
proc Serv_dccgevent {index st out} {
	global tl MVS

	set xc 0
	set in $MVS(DCC.$index.sock)
	set leng $MVS(DCC.$index.size)
	uplevel #0 set MVS(DCC.$index.time) [clock seconds]
	set fail_type 0
	if {[eof $in]} {
		if {$tl($in) < $leng} {
			set msg "Transfer Interrupted"
			set fail_type 0
		} elseif {$tl($in) > $leng} {
			set msg "Too much data transferred!!"
			set fail_type 0
		} else {
			set sx s
			if {[set st [expr {[clock seconds] - $st}]] == 0} {
				set st 1
				set sx {}
			}
			set xc 1
			set msg "Transfer completed. [expr {$leng / ($st * 1024.0)}] Kbytes/sec"
			set fail_type 1
		}
		unset sx st
	} {
		if {![catch {set buffer [read $in]} msg]} {
			incr tl($in) [set l [string length $buffer]]
			LogIt "downloaded $l bytes ($tl($in) total)"
			if {[set dt [expr {[clock seconds] - $st}]] == 0 || $tl($in) == 0} {
				set elt 0
			} {
				set elt [expr {($leng - $tl($in)) / ($tl($in) /([clock seconds] - $st))}]
			}
			if {$leng == 0} {
				set xt 0
			} {
				set xt [expr {($tl($in) * 100.0) / $leng}]
			}
		
			if {![catch {puts -nonewline $out $buffer} msg]} {
				if {![catch {puts -nonewline $in [binary format I1 $tl($in)]} msg]} {
					flush $in
					unset xc in leng fail_type dt elt xt
					return
				}
			} else {
				set fail_type 0
			}
		} else {
			set fail_type 0
		}
    	}
	catch {close $out} 
	Serv_endDCC Get $index $fail_type $msg
	catch {unset xc in leng fail_type l dt elt xt}
}

#
# End a DCC Transfer
#
# Usage: Serv_endDCC transfer_type index fail_type debug_info
#
# This function will end a user's DCC transfer. If it is a failure, it
# will print information about the transfer out to the log, if it is
# successful and the transfer is an incoming avatar, it will announce the
# avatar to all connected users. Once it's done ending the transfer, it
# will cleanup the mess made.
#
# checked for maximum memmory efficiency 12/10/199 22:58
#
proc Serv_endDCC {type index fail_type debug} {
        global MVS tl

        if !$fail_type {
                LogIt "($MVS(DCC.$index.sender) (DCC$type) - $debug"
        }
	catch {close $MVS(DCC.$index.sock)}
	set idx [lsearch -exact $MVS(dcc_list) $index]
	set MVS(dcc_list) [lreplace $MVS(dcc_list) $idx $idx]
	if {$type == "Get" && $fail_type} {
		if [CheckGif "$MVS(DCC.$index.file)"] {
			SendToAllUsers "AVATAR $MVS($MVS(DCC.$index.sender).name) [file tail $MVS(DCC.$index.file)] $MVS(DCC.$index.av_head_x) $MVS(DCC.$index.av_head_y) $MVS(DCC.$index.size) $MVS(DCC.$index.av_baloon_x) $MVS(DCC.$index.av_baloon_y)"
		} else {
			set MVS($MVS(DCC.$index.sender).avatar) "default.gif"
			SendToAllUsers "AVATAR $MVS($MVS(DCC.$index.sender).name) default.gif 0 20 [file size $MVS(icons)/default.gif] 24 6"
			SendToUser $MVS(DCC.$index.sender) "TOOBIG"
		}
	}
	catch {
		unset idx tl($MVS(DCC.$index.sock))
		unset MVS(DCC.$index.time) MVS(DCC.$index.server) \
		MVS(DCC.$index.file) MVS(DCC.$index.sender) MVS(DCC.$index.size) \
		MVS(DCC.$index.posn) MVS(DCC.$index.server) MVS(DCC.$index.sock) \
		MVS(DCC.$index.port) MVS(DCC.$index.remote) \
		MVS(DCC.$index.av_head_x) MVS(DCC.$index.av_head_y) \
		MVS(DCC.$index.av_baloon_x) MVS(DCC.$index.av_baloon_y)
	}
}

#
# Check gif file size
#
# Usage: CheckGif "filename"
#
# This will read in the GIF File header and determine it's size. It will
# comare the size against the server limits and report back if the file is
# good or bad.
#
# checked for maximum memmory efficiency 12/10/1999 23:00
#
proc CheckGif {file} {
	global MVS

	set infile [open $file r]
	fconfigure $infile -translation binary
	set bits [read $infile 10]
	close $infile
	if {[string range $bits 0 2] != "GIF"} {
		LogIt "(CheckGif) FAILED! NOT A GIF"
		unset infile bits
		return 0
	}
	binary scan $bits s* var
	if {[lindex $var 4] <= $MVS(maxheight) && [lindex $var 3] <= $MVS(maxwidth)} {
		LogIt "(CheckGif) PASSED! [lindex $var 3] X [lindex $var 4]"
		unset infile bits var
		return 1
	} else {
		LogIt "(CheckGif) FAILED! [lindex $var 3] X [lindex $var 4]"
		unset infile bits var
		return 0
	}
}

#
# Send objects to a user.
#
# Usage: SendObjects socket
#
# This function will send all server objects to a user. It is normally
# used at connect time to initalize an object for a user. 
#
# checked for maximum memmory efficiency 12/10/1999 23:03
#
proc SendObjects {who} {
	global MVS
	foreach object [glob -nocomplain "$MVS(sobjects)/*"] {
		source $object
	}	
}

Serv_CheckTimeouts

if $MVS(standalone) {
	vwait MVS(waiter)
}

